import hashlib


def index(tokens, options):
    """
    Given a list of tokens indexes the whole document into a set of them
    :param tokens:
    :return:
    """
    k = options.get("k_value", 50)
    window_size = options.get("window_size_value", 20)

    k_grams = generate_n_grams(tokens, k)
    hash_list = [hash_k_gram(k_gram) for k_gram in k_grams]
    windows = generate_n_grams(hash_list, window_size)
    fingerprints = select_fingerprints(windows)

    return fingerprints


def generate_n_grams(items, n):
    n_grams = []
    for i in range(len(items) - n + 1):
        n_gram = tuple(items[i:i + n])
        n_grams.append(n_gram)
    return n_grams


def hash_k_gram(k_gram):
    start_line = k_gram[0][1]
    end_line = k_gram[-1][1]
    concat_tokens = " ".join(map(lambda x: x[0], k_gram))
    hash_value = int(hashlib.sha512(concat_tokens.encode()).hexdigest(), 16)
    return hash_value, start_line, end_line


def select_fingerprints(windows):
    fingerprints = []
    previous_min_hash = None
    i = 0
    for window in windows:
        current_min_hash = min(window, key=lambda x: x[0])
        if current_min_hash != previous_min_hash:
            rindex = len(window) - window[-1::-1].index(current_min_hash) - 1
            fingerprint_position = i + rindex
            fingerprints.append((current_min_hash, fingerprint_position))
            previous_min_hash = current_min_hash
        i += 1
    return fingerprints



import pygments.token
from pygments import lexers

EMPTY_LINE = ""
NEW_LINE = '\n'

PYTHON_EXTENSION = "py"
C_EXTENSION = "c"
GO_EXTENSION = "go"

EXTENSION_NAME = "extension_name"
LEXER_NAME_KEY = "lexer_name"
SINGLE_COMMENT_STR_KEY = "single_comment"
SINGLE_COMMENT_SPECIAL_STR_KEY = "single_comments_special"
OPEN_MULTI_COMMENT_STR_KEY = "open_multi_comment"
CLOSE_MULTI_COMMENT_STR_KEY = "close_multi_comment"

_extension_map = {
    PYTHON_EXTENSION: {
        EXTENSION_NAME: PYTHON_EXTENSION,
        LEXER_NAME_KEY: "python",
        SINGLE_COMMENT_STR_KEY: "#",
        SINGLE_COMMENT_SPECIAL_STR_KEY: "\"",
        OPEN_MULTI_COMMENT_STR_KEY: "\"\"\"",
        CLOSE_MULTI_COMMENT_STR_KEY: "\"\"\""
    },
    C_EXTENSION: {
        EXTENSION_NAME: C_EXTENSION,
        LEXER_NAME_KEY: "c",
        SINGLE_COMMENT_STR_KEY: "//",
        OPEN_MULTI_COMMENT_STR_KEY: "/*",
        CLOSE_MULTI_COMMENT_STR_KEY: "*/"
    },
    GO_EXTENSION: {
        EXTENSION_NAME: GO_EXTENSION,
        LEXER_NAME_KEY: "go",
        SINGLE_COMMENT_STR_KEY: "//",
        OPEN_MULTI_COMMENT_STR_KEY: "/*",
        CLOSE_MULTI_COMMENT_STR_KEY: "*/"
    }
}


def process_tokens(tokens):
    processed_tokens = []
    line_count = 1
    for token in tokens:
        token_value, raw_code = token
        processed_tokens.append((token_value, line_count, raw_code))
        line_count += raw_code.count(NEW_LINE)
    return processed_tokens


def filter_tokens(tokens):
    stop_tokens = [pygments.token.Text, pygments.token.Comment,
                   pygments.token.Literal.String.Doc, pygments.token.Comment.Single, pygments.token.Literal.String.Doc]
    return list(filter(lambda token: token[0] not in stop_tokens, tokens))


def format_tokens(tokens):
    return list(map(lambda token: (str(token[0]), token[1]), tokens))


def preprocess(raw_code, options):
    """
    Preprocess a given raw code.
    :param raw_code: Raw source code to be preprocessed
    :return: A filtered list in which each item is a tuple with the following data (token, line)
    """
    extension = _extension_map.get(options.get("extension"))
    lexer_name = extension.get(LEXER_NAME_KEY)
    lexer = lexers.get_lexer_by_name(lexer_name)
    tokens = process_tokens(lexer.get_tokens(raw_code))
    print("raw_tokens", tokens)
    filtered_tokens = filter_tokens(tokens)

    return format_tokens(filtered_tokens)



import logging
import uuid

from .config import fingerprints_cache
from .indexers.winnowing_indexer import index
from .preprocessors.preprocesser import preprocess
from .searchers.exact_searcher import search

logger = logging.getLogger(__name__)

K_VALUE = 15
W_VALUE = 10


def calculate_file_lines(file_content):
    lines = file_content.split('\n')
    return len(lines)


def calculate_file_matched_lines(matches):
    matched_lines = 0
    for match in matches:
        _, start_line, end_line = match
        matched_lines += end_line - start_line + 1
    return matched_lines


def calculate_match_percentage(file_content, matches):
    file_lines = calculate_file_lines(file_content)
    matched_lines = calculate_file_matched_lines(matches)
    return matched_lines / file_lines


def get_fingerprints_2(raw_code):

    index_options = {
        "k_value": K_VALUE,
        "window_size_value": W_VALUE
    }

    preprocessed_file = preprocess(
        raw_code, {"extension": "py"})
    print("preprocessed_file", preprocessed_file)
    fingerprints = index(preprocessed_file, index_options)
    print("fingerprints", fingerprints)

    return fingerprints


def get_fingerprints(file):

    index_options = {
        "k_value": K_VALUE,
        "window_size_value": W_VALUE
    }

    preprocess_options = {
        "extension": file.extension
    }

    cached_fingerprints = fingerprints_cache.get(file.id)

    if cached_fingerprints:
        return cached_fingerprints

    preprocessed_file = preprocess(
        file.content, preprocess_options)
    fingerprints = index(preprocessed_file, index_options)
    fingerprints_cache[file.id] = fingerprints

    return fingerprints


def build_result(file, matches):
    match_percentage = calculate_match_percentage(file.content, matches)
    total_lines_matched = calculate_file_matched_lines(matches)

    return {
        "id": file.id,
        "path": file.path,
        "name": file.name,
        "extension": file.extension,
        "match_percentage": match_percentage,
        "total_lines_matched": total_lines_matched,
        "lines_matched": matches
    }


def compare_files(source_file, target_file, _):

    logger.info("Comparing files (id, name): '({}, {})' vs. '({},{})'".format(
        source_file.id, source_file.name, target_file.id, target_file.name))

    source_fingerprints = get_fingerprints(source_file)
    target_fingerprints = get_fingerprints(target_file)

    source_matches, target_matches = search(
        source_fingerprints, target_fingerprints)

    source_file_results = build_result(source_file, source_matches)
    target_file_results = build_result(target_file, target_matches)

    return {
        "id": str(uuid.uuid4()),
        "source_file": source_file_results,
        "target_file": target_file_results
    }
